// Copyright Matthias Troyer
// 2005. Distributed under the Boost Software License, Version
// 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/timer.hpp>
#include <boost/iostreams/stream_buffer.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/iostreams/device/back_inserter.hpp>
#include <boost/serialization/vector.hpp>
#include <vector>
#include <iostream>
#include <string>
#include <sstream>

namespace test
{

  int live_code;
  
  namespace io = boost::iostreams;
  
  typedef std::vector<char> buffer_type;
  
  typedef io::stream< io::back_insert_device <buffer_type> > my_ostream;
  typedef io::stream< io::basic_array_source <char> > my_istream;
      
  buffer_type buffer1;
  buffer_type buffer2;

  double write_serialization(std::size_t number_of_elements)
  {
      const std::vector<double> data(number_of_elements);
      my_ostream output_stream(buffer1);
      boost::archive::binary_oarchive oa(output_stream);

      std::size_t s = data.size();
      buffer1.reserve(10240+sizeof(std::size_t)+number_of_elements*sizeof(double));
      
      boost::timer time;
      oa << data;
      output_stream.flush();
      live_code=buffer1.size();
      return time.elapsed();
  }

  double write_direct(std::size_t number_of_elements)
  {
      std::vector<double> data(number_of_elements);
      
      
      std::size_t s = data.size();
      buffer2.reserve(sizeof(std::size_t)+number_of_elements*sizeof(double));

      boost::timer time;
      char const * p = reinterpret_cast<char*>(&s);
      buffer2.insert(buffer2.end(),p,p+sizeof(std::size_t));
      
      p = reinterpret_cast<char*>(&(data[0]));
      buffer2.insert(buffer2.end(),p,p+number_of_elements*sizeof(double));
      
      live_code=buffer2.size();
      
      return time.elapsed();
  }
  

  double read_serialization()
  {
      std::vector<double> data;
      
      io::basic_array_source <char> source(&(buffer1[0]),buffer1.size());
      my_istream input_stream(source);
      boost::archive::binary_iarchive ia(input_stream);
      
      boost::timer time;
      ia >> data;
      input_stream.close();
      
      live_code += data.size();
      return time.elapsed();
  }

  double read_memcpy()
  {
      std::vector<double> data;
      
      boost::timer time;
      std::size_t s;
      
      s = *reinterpret_cast<std::size_t*>(&(buffer2[0]));

      data.resize(s);
      
      double *p = reinterpret_cast<double*>(&buffer2[sizeof(std::size_t)]);
      std::memcpy(&data[0],p,s*sizeof(double));

      live_code += data.size();
      return time.elapsed();
  }

  double read_copy()
  {
      std::vector<double> data;
      
      boost::timer time;
      std::size_t s;
      
      s = *reinterpret_cast<std::size_t*>(&(buffer2[0]));

      data.reserve(s);
      
      double *p = reinterpret_cast<double*>(&buffer2[sizeof(std::size_t)]);
      std::copy(p,p+s,std::back_inserter(data));
      live_code += data.size();
      return time.elapsed();
  }


}


int main()
{    

    const std::size_t number_elements=10000000;
    std::cout
        << "serialization writing time:           "
        << test::write_serialization(number_elements)
        << std::endl;
    
    std::cout
        << "direct writing time: "
        << test::write_direct(number_elements)
        << std::endl;


    std::cout
        << "serialization reading time:           "
        << test::read_serialization()
        << std::endl;
    
    std::cout
        << "std::memcpy reading time: "
        << test::read_memcpy()
        << std::endl;

    std::cout
        << "std::copy reading time: "
        << test::read_copy()
        << std::endl;


    // This is ultimately responsible for preventing all the test code
    // from being optimized away.  Change this to return 0 and you
    // unplug the whole test's life support system.
    return test::live_code < 0;
}

